import jittor as jt
from jittor.dataset import *
from PIL import Image
import numpy as np
import unittest



class TestSamplerDataset(Dataset):
    def __init__(self):
        super().__init__()
        self.set_attrs(total_len=40, batch_size=1)

    def __getitem__(self, idx):
        return idx**2


class TestSampler(unittest.TestCase):
    def test_sequential_sampler(self):
        testdataset = TestSamplerDataset()
        seqsampler = SequentialSampler(testdataset)
        assert len(seqsampler) == 40
        for idx, batch in enumerate(seqsampler):
            assert idx == batch
        for i, data in enumerate(testdataset):
            assert data.item() == i**2

    def test_random_sampler(self):
        testdataset = TestSamplerDataset()
        randomsampler = RandomSampler(testdataset)
        assert len(randomsampler) == 40
        diff = 0
        for i, data in enumerate(testdataset):
            diff += data.item() == i**2
        assert diff < 10

    def test_subset_random_sampler(self):
        testdataset = TestSamplerDataset()
        subsetsampler = SubsetRandomSampler(testdataset, (20, 30))
        assert len(subsetsampler) == 10
        s = 0
        for i, data in enumerate(testdataset):
            s += data.item()
        s2 = 0 
        for i in range(20,30):
            s2 += i**2
        assert s == s2, (s, s2)

    def test_batch_sampler(self):
        testdataset = TestSamplerDataset()
        seqforbatch = SequentialSampler(testdataset)
        batchsampler = BatchSampler(seqforbatch, 4, drop_last=False)
        assert len(batchsampler) == 10
        for batch in batchsampler:
            assert len(batch) == 4


if __name__ == "__main__":
    unittest.main()



# ***************************************************************
# Copyright (c) 2022 Jittor. All Rights Reserved. 
# Maintainers:
#   Dun Liang <randonlang@gmail.com>.
#   Xiangli Li <190569238@qq.com>
#
# 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************

import jittor as jt
import numpy as np

class SparseVar:
    def __init__(self,indices,values,shape):
        assert isinstance(indices,jt.Var) and isinstance(values,jt.Var) and isinstance(shape,jt.NanoVector)
        self.indices = indices
        self.values = values
        self.shape = shape
        self.ndim = len(shape)
        
    def _indices(self):
        return self.indices
    
    def _values(self):
        return self.values
    
    def t(self):
        indices = list(self.indices.split(1,dim=0))
        indices[-1],indices[-2] = indices[-2],indices[-1]
        indices = jt.concat(indices,dim=0)
        shape = list(self.shape)
        shape[-1],shape[-2] = shape[-2],shape[-1]
        shape = jt.NanoVector(shape)
        return SparseVar(indices,self.values,shape)
        
    def to_dense(self):
        ret = jt.zeros(self.shape,self.values.dtype)
        indices  = tuple(self.indices.split(1,dim=0))
        ret[indices]=self.values
        return ret

def sparse_array(indices,values,shape):
    return SparseVar(indices,values,shape)

def spmm(spase_x,y):
    assert isinstance(spase_x,SparseVar) and isinstance(y,jt.Var)
    assert spase_x.ndim==2 and y.ndim==2 and spase_x.shape[-1]==y.shape[0]
    
    # TODO
    x = spase_x.to_dense()
    return jt.matmul(x,y)
    


# ***************************************************************
# Copyright (c) 2022 Jittor. All Rights Reserved. 
# Maintainers: Dun Liang <randonlang@gmail.com>. 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np
from jittor.utils.pytorch_converter import convert
import os

try:
    jt.dirty_fix_pytorch_runtime_error()
    import torch
    from torch import nn
except:
    torch = None

code="""
import torch.nn as nn
import torch.utils.model_zoo as model_zoo 

__all__ = ['ResNet', 'resnet18', 'resnet34', 'resnet50', 'resnet101', 'resnet152']
 
 
model_urls = {
    'resnet18': 'https://download.pytorch.org/models/resnet18-5c106cde.pth',
    'resnet34': 'https://download.pytorch.org/models/resnet34-333f7ec4.pth',
    'resnet50': 'https://download.pytorch.org/models/resnet50-19c8e357.pth',
    'resnet101': 'https://download.pytorch.org/models/resnet101-5d3b4d8f.pth',
    'resnet152': 'https://download.pytorch.org/models/resnet152-b121ed2d.pth',
}
 
 
def conv3x3(in_planes, out_planes, stride=1):
    '''3x3 convolution with padding'''
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride, padding=1, bias=False)
    
class BasicBlock(nn.Module):
    expansion = 1
 
    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU()
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = nn.BatchNorm2d(planes)
        self.downsample = downsample
        self.stride = stride
 
    def forward(self, x):
        residual = x
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)
        out = self.conv2(out)
        out = self.bn2(out)
        if self.downsample is not None:
            residual = self.downsample(x)
        out += residual
        out = self.relu(out)
        return out
        
class Bottleneck(nn.Module):
    expansion = 4
 
    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride,
                               padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, planes * self.expansion, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(planes * self.expansion)
        self.relu = nn.ReLU()
        self.downsample = downsample
        self.stride = stride
 
    def forward(self, x):
        residual = x
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)
        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)
        out = self.conv3(out)
        out = self.bn3(out)
        if self.downsample is not None:
            residual = self.downsample(x)
        out += residual
        out = self.relu(out)
        return out
        
        
class ResNet(nn.Module):
 
    def __init__(self, block, layers, num_classes=1000):
        self.inplanes = 64
        super(ResNet, self).__init__()
        self.conv1 = nn.Conv2d(3, 64, kernel_size=7, stride=2, padding=3,
                               bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.relu = nn.ReLU()
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride=2)
        self.layer3 = self._make_layer(block, 256, layers[2], stride=2)
        self.layer4 = self._make_layer(block, 512, layers[3], stride=2)
        self.avgpool = nn.AvgPool2d(7, stride=1)
        self.fc = nn.Linear(512 * block.expansion, num_classes)
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out')
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
 
    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(planes * block.expansion),
            )
        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes))
 
        return nn.Sequential(*layers)
 
    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        x = self.avgpool(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)
 
        return x
        
def resnet18(pretrained=False, **kwargs):
    '''Constructs a ResNet-18 model.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    '''
    model = ResNet(BasicBlock, [2, 2, 2, 2], **kwargs)
    if pretrained:
        model.load_state_dict(model_zoo.load_url(model_urls['resnet18']))
    return model
 
 
def resnet34(pretrained=False, **kwargs):
    '''Constructs a ResNet-34 model.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    '''
    model = ResNet(BasicBlock, [3, 4, 6, 3], **kwargs)
    if pretrained:
        model.load_state_dict(model_zoo.load_url(model_urls['resnet34']))
    return model
 
 
def resnet50(pretrained=False, **kwargs):
    '''Constructs a ResNet-50 model.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    '''
    model = ResNet(Bottleneck, [3, 4, 6, 3], **kwargs)
    if pretrained:
        model.load_state_dict(model_zoo.load_url(model_urls['resnet50']))
    return model
 
 
def resnet101(pretrained=False, **kwargs):
    '''Constructs a ResNet-101 model.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    '''
    model = ResNet(Bottleneck, [3, 4, 23, 3], **kwargs)
    if pretrained:
        model.load_state_dict(model_zoo.load_url(model_urls['resnet101']))
    return model
 
 
def resnet152(pretrained=False, **kwargs):
    '''Constructs a ResNet-152 model.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
    '''
    model = ResNet(Bottleneck, [3, 8, 36, 3], **kwargs)
    if pretrained:
        model.load_state_dict(model_zoo.load_url(model_urls['resnet152']))
    return model
"""

@unittest.skipIf(torch is None, "pytorch not found.")
class TestPytorchConverter(unittest.TestCase):
    def test_pytorch_converter(self):
        name1 = os.path.join(jt.flags.cache_path, 'test_pytorch_converter_1.py')
        print(f"save source code into {name1}")
        with open(name1, 'w') as f:
            f.write(code)
        
        ret = convert(code)

        name2 = os.path.join(jt.flags.cache_path, 'test_pytorch_converter_2.py')
        print(f"save destination code into {name2}")
        with open(name2, 'w') as f:
            f.write(ret)
        
        from test_pytorch_converter_1 import resnet18 as torch_resnet18
        from test_pytorch_converter_2 import resnet18 as jittor_resnet18
        model_torch = torch_resnet18(False)
        model_jittor = jittor_resnet18(False)
        model_jittor.load_parameters(model_torch.state_dict())

        img = np.random.randn(1,3,224,224).astype("float32")
        img_torch = torch.Tensor(img)
        img_jittor = jt.array(img)

        out_torch = model_torch(img_torch)
        out_jittor = model_jittor(img_jittor)
        assert abs((out_torch.cpu().detach().numpy() - out_jittor.data)).mean() < 1e-4

if __name__ == "__main__":
    unittest.main()



# ***************************************************************
# Copyright (c) 2022 Jittor. All Rights Reserved. 
# Maintainers: Dun Liang <randonlang@gmail.com>. 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import os

@unittest.skipIf(not jt.compile_extern.use_mkl, "Not use mkl, Skip")
class TestMklTestOp(unittest.TestCase):
    def test(self):
        assert jt.mkl_ops.mkl_test().data==123

if __name__ == "__main__":
    unittest.main()



import jittor as jt
from jittor import nn

def can_softmax_v1(a, dim):
    if not jt.flags.use_cuda:
        return False
    if dim != -1 and dim != len(a.shape)-1:
        return False
    if a.shape[len(a.shape)-1] > 10000:
        return False
    return True

def softmax_v1(a, log=False):
    assert can_softmax_v1(a, -1)
    length = a.shape[-1]
    # tnum = 1024
    tnum = 500 if length % 500 == 0 else 512
    tnum = 125 if length % 125 == 0 else 128
    # tnum = 125
    # tnum = 1000 if length % 1000 == 0 else 1024
    # tnum = 250
    per_thread = (length-1) // tnum + 1
    ILP = 1
    for ilp in [8,4,2]:
        if length % tnum == 0 and per_thread % ilp == 0:
            ILP = ilp
            per_thread //= ILP
            break
    for_loop = f"""
    #pragma unroll
    for (int i=0; i<{per_thread}; i++)
    """
    if length % tnum != 0:
        for_loop += f"if ((i*{tnum}+threadIdx.x)*{ILP} < len)\n"

    return jt.code(a.shape, a.dtype, [a], cuda_header=f'''
#include <{jt.compile_extern.cub_home}cub/cub.cuh>
#include <type/fp16_compute.h>
''', cuda_src=f'''
__global__ void kernel(in0_type* x, out0_type* y, int len) {{
    typedef cub::BlockReduce<float, {tnum}> BlockReduce;
    __shared__ typename BlockReduce::TempStorage temp_storage;

    int id = blockIdx.x * len;
    in0_type v[{per_thread}][{ILP}];
    {for_loop}
        vload<sizeof(in0_type)*{ILP}>(v[i], &x[id+(i*{tnum}+threadIdx.x)*{ILP}]);
    // v[i] = x[id+i*{tnum}+threadIdx.x];
    float v1 = -1e30;
    {for_loop}
        #pragma unroll
        for (int j=0; j<{ILP}; j++) {{
            v1 = max(v1, float(v[i][j]));
        }}
    __shared__ float vmax;
    auto tmp = BlockReduce(temp_storage).Reduce(v1, cub::Max());
    if (threadIdx.x == 0)
        vmax = tmp;
    __syncthreads();

    v1 = 0;
    {for_loop}
        #pragma unroll
        for (int j=0; j<{ILP}; j++) {{
            v[i][j] = expf(float(v[i][j]) - vmax);
            v1 += float(v[i][j]);
        }}

    tmp = BlockReduce(temp_storage).Sum(v1);
    __shared__ float vsum;
    if (threadIdx.x == 0)
        vsum = tmp;
    __syncthreads();

    {for_loop}
        #pragma unroll
        for (int j=0; j<{ILP}; j++)
            v[i][j] = {
                "@expand_op(log,@in0_type,float(v[i][j])/vsum)" if log
                else "float(v[i][j])/vsum"
            };
    {for_loop}
        vload<sizeof(in0_type)*{ILP}>(&y[id+(i*{tnum}+threadIdx.x)*{ILP}], v[i]);
}}
int len = in0->shape[in0->shape.size()-1];
int bnum = in0->numel() / len;
cudaGetLastError();
kernel<<<bnum, {tnum}>>>(in0_p, out0_p, len);
CHECK(0 == cudaGetLastError());
''', cuda_grad_src=[f"""
__global__ void kernel(pout0_type* x, dout_type* y, out0_type* z, int len) {{
    int id = blockIdx.x * len;
    in0_type vx[{per_thread}][{ILP}];
    in0_type vy[{per_thread}][{ILP}];
    {for_loop} {{
        vload<sizeof(in0_type)*{ILP}>(vx[i], &x[id+(i*{tnum}+threadIdx.x)*{ILP}]);
        vload<sizeof(in0_type)*{ILP}>(vy[i], &y[id+(i*{tnum}+threadIdx.x)*{ILP}]);
    }}
    float v1 = 0;
    {for_loop} 
        #pragma unroll
        for (int j=0; j<{ILP}; j++)
            v1 += {"float(vy[i][j]);" if log else "float(vx[i][j]*vy[i][j]);"}

    typedef cub::BlockReduce<float, {tnum}> BlockReduce;
    __shared__ typename BlockReduce::TempStorage temp_storage;
    auto tmp = BlockReduce(temp_storage).Sum(v1);
    __shared__ float reduce_var;
    if (threadIdx.x == 0)
        reduce_var = tmp;
    __syncthreads();

    {for_loop}
        #pragma unroll
        for (int j=0; j<{ILP}; j++)
            vx[i][j] = {
                "vy[i][j] - in0_type(expf(vx[i][j]) * reduce_var);" if log 
                else "vx[i][j] * (vy[i][j] - in0_type(reduce_var));"
            }

    {for_loop}
        vload<sizeof(in0_type)*{ILP}>(&z[id+(i*{tnum}+threadIdx.x)*{ILP}],
            vx[i]);
}}
int len = in0->shape[in0->shape.size()-1];
int bnum = in0->numel() / len;
cudaGetLastError();
kernel<<<bnum, {tnum}>>>(pout0_p, dout_p, out0_p, len);
CHECK(0 == cudaGetLastError());
"""])


# ***************************************************************
# Copyright (c) 2022 Jittor. All Rights Reserved. 
# Maintainers: Dun Liang <randonlang@gmail.com>. 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np
from .test_core import expect_error

class TestContrib(unittest.TestCase):
    def test_concat(self):
        def check(shape, dim, n):
            num = np.prod(shape)
            arr1 = []
            arr2 = []
            for i in range(n):
                a = (np.array(range(num)) + i*num).reshape(shape)
                arr1.append(a)
                arr2.append(jt.array(a))
            x = np.concatenate(tuple(arr1), dim)
            y = jt.concat(arr2, dim)
            assert (x==y.data).all(), (x, y.data, arr1, arr2)
        check([2,3,4], 0, 2)
        check([2,3,4], 1, 3)
        check([2,3,4], 2, 4)
        check([2,3,4,5], 0, 4)
        check([2,3,4,5], 2, 4)
        check([2,3,4,5], 3, 4)
        check([1], 0, 20)

    def test_slice(self):
        def check(shape, slices):
            x = jt.random(shape)
            a = x[slices].data
            b = x.data[slices]
            assert (a==b).all(), (a, b)
            y = x.numpy()
            v = jt.random(a.shape)
            x[slices] = v
            y[slices] = v.data
            assert (x.data==y).all()
        # TODO: when slice same row/col many times and assign value, numpy will retain the last value but we assign their sum. eg: check([3,3,3,3], ([[0,1,1]],slice(None),[[1],[2],[0]],1))
        check([3], (1))
        check([3,3,3,3], ([[0],[1]],slice(None),[1,2],1))
        check([3,3,3,3], (slice(None),slice(None),slice(None),slice(None)))
        check([3,3,3,3], ([0,1],[0,1],[0,1],[0,1]))
        check([3,3,3,3], ([0,1],-2,slice(None),[0,1]))
        check([3,3,3,3], ([0,1],slice(1,2,2),[1,2],1))
        check([3,3,3,3], ([0,1],slice(None),[1,2],1))
        check([10,10,10,10], (slice(1,None,2),slice(-1,None,2),[1,2],-4))
        check([20], 0)
        check([20], 10)
        check([20], -10)
        check([10,10,10,10], 1)
        check([10,10,10,10], (1,slice(None),2))
        check([10,10,10,10], (-2,slice(None),2,slice(1,9,2)))


if __name__ == "__main__":
    unittest.main()



# ***************************************************************
# Copyright (c) 2022 Jittor. All Rights Reserved. 
# Maintainers: Dun Liang <randonlang@gmail.com>. 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import unittest
import jittor as jt
import numpy as np
from .test_grad import ngrad
from .test_cuda import test_cuda

def check(op, *args):
    x = eval(f"np.{op}(*args)")
    y = eval(f"jt.{op}(*args).data")
    convert = lambda x: x.astype("uint8") if x.dtype=="bool" else x
    x = convert(x)
    y = convert(y)
    # str match nan and inf
    assert x.dtype == y.dtype and x.shape == y.shape, \
        (x.dtype, y.dtype, x.shape, y.shape)
    for a,b in zip(x.flatten(), y.flatten()):
        assert str(a)[:5] == str(b)[:5], (a,b)

class TestUnaryOp(unittest.TestCase):
    def test_unary_op(self):
        assert jt.float64(1).data.dtype == "float64"
        assert (jt.abs(-1) == 1).data.all()
        assert (abs(-jt.float64(1)) == 1).data.all()
        a = np.array([-1,2,3,0])
        check("abs", a)
        check("negative", a)
        check("logical_not", a)
        check("bitwise_not", a)
        b = np.array([1.1, 2.2, 3.3, 4.4, -1, 0])
        type = "float16" if (jt.flags.amp_reg & 2) else "float32"
        check("log", a.astype(type))
        check("exp", a.astype(type))
        check("sqrt", a.astype(type))
        
    def test_grad(self):
        ops = ["abs", "negative", "log", "exp", "sqrt",
            "sin", "arcsin", "sinh", "arcsinh", 
            "tan", "arctan", "tanh", "arctanh", 
            "cos", "arccos", "cosh", "arccosh", 
            "sigmoid", 
        ]
        a = np.array([1.1, 2.2, 3.3, 4.4])
        for op in ops:
            if op == "abs":
                b = np.array(a+[-1,])
            elif op == "arccosh":
                b = np.array(a)
            elif "sin" in op or "cos" in op or "tan" in op:
                b = np.array(a) / 5
            else:
                b = np.array(a)
            func = lambda x: eval(f"np.{op}(x[0]).sum()")
            if op == "sigmoid":
                func = lambda x: (1/(1+np.exp(-x[0]))).sum()
            x, (da,) = ngrad(func, [b], 1e-8)
            ja = jt.array(b)
            jb = eval(f"jt.{op}(ja)")
            jda = jt.grad(jb, ja)
            tol = 1e-2 if jt.flags.amp_reg & 2 else 1e-6
            assert (np.allclose(jda.data, da, atol=tol, rtol=tol)), (jda.data,da,op)

    def test_sigmoid(self):
        a = np.arange(-150,150, 10).astype("float32")
        # a = np.array([-150.0, -140.0, -130.0]).astype("float32")
        b = jt.array(a, dtype='float32')
        b1 = b.sigmoid().numpy()
        assert np.isnan(b1).any() == False

    def test_safe_clip(self):
        a = jt.array([-1.0,0,0.4,1,2,3])
        b = a.safe_clip(0.1, 0.5)
        assert np.allclose(b.data, [0.1,0.1,0.4,0.5,0.5,0.5])
        da = jt.grad(b, a)
        assert (da.data == 1).all()

    def test_erfinv(self):
        from scipy import special
        y = np.linspace(-1.0, 1.0, num=10)
        x = special.erfinv(y)
        y2 = jt.array(y)
        x2 = jt.erfinv(y2)
        np.testing.assert_allclose(y.data, y2.data)
        
        
        y = np.linspace(-0.9, 0.9, num=10)
        x = special.erfinv(y)
        y2 = jt.array(y)
        x2 = jt.erfinv(y2)
        np.testing.assert_allclose(y.data, y2.data)
        d = jt.grad(x2, y2)
        _, (dn,) = ngrad(lambda y: special.erfinv(y).sum(), [y], 1e-8)
        tol = 1e-3 if jt.flags.amp_reg & 2 else 1e-6
        np.testing.assert_allclose(d.data, dn, atol=tol, rtol=tol)


class TestUnaryOpCuda(TestUnaryOp, test_cuda(2)):
    pass

class TestUnaryOpCpuFp16(TestUnaryOp, test_cuda(2)):
    def setUp(self):
        jt.flags.amp_reg = 2 | 4 | 8 | 16
    def tearDown(self):
        jt.flags.amp_reg = 0

class TestUnaryOpCudaFp16(TestUnaryOp, test_cuda(2)):
    def setUp(self):
        jt.flags.amp_reg = 2 | 4 | 8 | 16
        jt.flags.use_cuda = 1
    def tearDown(self):
        jt.flags.amp_reg = 0
        jt.flags.use_cuda = 0

if __name__ == "__main__":
    unittest.main()


# ***************************************************************
# Copyright (c) 2022 Jittor. All Rights Reserved. 
# Maintainers: Dun Liang <randonlang@gmail.com>. 
# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
# ***************************************************************
import os
os.environ['OPENBLAS_NUM_THREADS'] = '1'

import unittest
import time
import jittor as jt
from jittor import LOG
import numpy as np
from .test_core import expect_error
import contextlib

performance_test = os.environ.get("performance_test", "") == "1"
skip_slow_test = not performance_test

@contextlib.contextmanager
def performance_test_scope(warmup=0, rerun=0, **args):
    """ profile scope
    example:
        with jt.profile_scope() as report:
            ......
        print(report)
    """
    assert not jt.flags.profiler_enable
    if skip_slow_test:
        jt.profiler.start(0, 0)
    else:
        jt.profiler.start(warmup, rerun)
    report = []
    try:
        with jt.flag_scope(**args):
            yield report
    finally:
        jt.profiler.stop()
        if skip_slow_test:
            report.extend([[1e30]]*3)
        else:
            report.extend(jt.profiler.report())

def retry(num):
    def outer(func):
        def inner(*args):
            for i in range(num):
                if i == num-1:
                    func(*args)
                    break
                try:
                    func(*args)
                    break
                except:
                    pass
                LOG.v(f"Retry {i}")
        return inner
    return outer

def get_np_matmul_toughtput(size):
    # import os
    # os.environ['OPENBLAS_NUM_THREADS'] = '1'
    # import numpy as np
    # import time
    a = np.random.randn(size, size).astype("float32")
    b = np.random.randn(size, size).astype("float32")
    c = np.random.randn(size, size).astype("float32")
    warmup = 2
    rerun = 10+1
    for _ in range(warmup): np.matmul(a,b,c)
    start_time = time.time()
    for _ in range(rerun): np.matmul(a,b,c)
    elapsed_time = time.time() - start_time
    return (size*size*size*rerun) / elapsed_time

class TestFusedOp(unittest.TestCase):
    def test_add(self):
        jt.clean()
        def check(hv, lv, lo):
            self.assertEqual((
                jt.number_of_hold_vars(),
                jt.number_of_lived_vars(),
                jt.number_of_lived_ops()),
                (hv, lv, lo))
        for i in range(8):
            check(0,0,0)
            a = jt.array(1.0).name('a').stop_fuse()
            b = (a+jt.array(1.0).name('t1').stop_fuse()).name('b')
            c = (b+jt.array(1.0).name('t2').stop_fuse()).name('c')
            check(3,5,5)
            graph = jt.dump_all_graphs()
            # for n in graph.nodes_info:
            #     print(n)
            self.assertEqual(c.data, 3)
            graph2 = jt.dump_all_graphs()
            print("check", i)
            for n in graph2.nodes_info:
                print(n)
            print(jt.liveness_info())
            check(3,5,2)
            graph = jt.dump_all_graphs()
            for node in graph.nodes_info:
                if node.startswith("Op"):
                    if 'add->' in node:
                        assert ':s0' in node, node
                    else:
                        assert ':s1' in node, node
                elif ',b,' in node:
                    # b has been fused
                    assert ':s0' in node, node
                else:
                    assert ':s1' in node
            if i&1: del a
            if i&2: del b
            if i&4: del c
            
            if i==0: check(3,5,2)
            elif i==1: check(2,5,2)
            elif i==2: check(2,5,2)
            elif i==3: check(1,1,0)
            elif i==4: check(2,3,1)
            elif i==5: check(1,3,1)
            elif i==6: check(1,1,0)
            elif i==7: check(0,0,0)
            
            if not (i&1): a.sync()
            if not (i&2): b.sync()
            if not (i&4): c.sync()
            
            if i==0: check(3,5,2)
            elif i==1: check(2,3,1)
            elif i==2: check(2,5,2)
            elif i==3: check(1,1,0)
            elif i==4: check(2,3,1)
            elif i==5: check(1,1,0)
            elif i==6: check(1,1,0)
            
            if not (i&1): del a
            if not (i&2): del b
            if not (i&4): del c
            check(0,0,0)
            
    def test_fuse_reduce_and_broadcast(self):
        size = 10
        a = jt.random([size,size,1])
        b = jt.random([1,size,size])
        c = (a*b).sum(dim=1)
        nc = (a.data*b.data).sum(1)
        assert c.shape == [size,size]
        assert (np.abs(nc-c.data)<1e-5).all()
        
    def test_fuse_reduce_and_broadcast2(self):
        size = 10
        a = jt.random([1])
        b = a.broadcast([size]).sum()
        assert (np.abs(b.data - a.data*size) < 1e-5).all()
        a = jt.random([size,1])
        b = a.broadcast([size,size]).sum(1, keepdims=True)
        assert (np.abs(b.data - a.data*size) < 1e-5).all()
    
    def test_fuse_reduce2(self):
        size = 10
        a = jt.random([1]).broadcast([size]).name('a')
        # a.data
        b = a.sum().name('b')
        c = a.min().name('c')
        d = a.max().name('d')
        jt.fetch_sync([b,c,d])
        
        graph = jt.dump_all_graphs()
        node_a = [ node for node in graph.nodes_info if ",a," in node ]
        assert 's0' in node_a[0]
        
        v = a.data[0]
        assert np.allclose(v*10,b.data) and v==c.data and v==d.data, (v, b.data, c.data, d.data)
        
    def test_profile_fused_op(self):
        size = 1000
        r1 = []
        r2 = []
        for size in range(1024, 1025, 1):
            with performance_test_scope(2, 10) as report:
                a = jt.random([size,size,1])
                b = jt.random([1,size,size])
                c = (a*b).sum(1)
                c.sync()
            
            assert len(report) == 3
            tp_np = get_np_matmul_toughtput(size)
            tp_jt = float(report[1][-1])
            r1.append(tp_jt)
            r2.append(tp_np)
            na = a.data.reshape((size,size))
            nb = b.data.reshape((size,size))
            nc = np.matmul(na,nb)
            assert (np.abs(nc-c.data)<1e-2).all(), np.abs(nc-c.data).max()
    
    # @unittest.skipIf(skip_slow_test, "Skip slow test")
    def test_profile_fused_op_transpose(self):
        for size in range(1024, 1025, 1):
            with performance_test_scope(2, 10):
                b = jt.random([size,1,size])
                a = jt.random([1,size,size])
                c = (a*b).sum(2)
                c.data
        
    # @unittest.skipIf(skip_slow_test, "Skip slow test")
    def test_profile_fused_op_split(self):
        # match v4
        @retry(10)
        def check(n, m, k, cs, rs, rtp):
            a = jt.random([n,m,1])
            b = jt.random([1,m,k])
            a.data, b.data
            with performance_test_scope(
                20, 20000000000//(n*m*k),
                compile_options = {
                "split0":16,"split1":6,"split2":16,
                "order0":0, "order1":1,"order2":1,
                "order3":0, "order4":0,"order5":0,
                "restride":rs, "unroll":2, "vectorize":2,
                "compile_shapes":cs
            }) as report:
                c = (a*b).sum(1)
            c.data
            
            na = a.data.reshape((n,m))
            nb = b.data.reshape((m,k))
            nc = np.matmul(na,nb)
            
            assert (np.abs(nc-c.data)<1e-2).all(), np.abs(nc-c.data).max()
            tp = float(report[-1][-1])
            assert tp > rtp * 10**9, (tp, rtp)

        check(65, 8, 19, 1, 0, 0)
        check(64, 6, 16, 1, 0, 33) # TODO 36
        check(64, 6, 16, 0, 0, 21)
        check(64, 60, 16, 1, 0, 44)
        check(64, 60, 16, 0, 0, 30)
        check(65, 60, 16, 0, 0, 30)
        check(65, 61, 16, 0, 0, 27)
        check(65, 65, 16, 0, 0, 26)
        check(64, 60, 64, 1, 1, 27)
        check(64, 60, 64, 1, 0, 42)
        check(64, 60, 64, 0, 0, 30) # TODO: why slower?

    def test_array_reindex(self):
        a = jt.array([1])
        b = a.reindex([3], ['i0-1'])
        np.testing.assert_allclose(b.data, [0,1,0])

    
    @unittest.skipIf(skip_slow_test, "Skip slow test")
    def test_profile_fused_op_restride(self):
        # match v6

        @retry(10)
        def check(n, m, k, cs, rs, pa, rtp):
            a = jt.random([n,m,1])
            b = jt.random([1,m,k])
            a.data, b.data
            with performance_test_scope(
                0, 20000000000//(n*m*k),
                compile_options = {
                "order0":0, "order1":0,"order2":0,
                "split0":64,"split1":60,"split2":64,
                "split3":16,"split4":6,"split5":16,
                "order3":0, "order4":1,"order5":1,
                "order6":0, "order7":0,"order8":0,
                "restride":rs,"vectorize":2,"unroll":2,
                "compile_shapes":cs, "parallel":pa
            }) as report:
                c = (a*b).sum(1)
            c.sync()

            na = a.data.reshape((n,m))
            nb = b.data.reshape((m,k))
            nc = np.matmul(na,nb)
            assert (np.abs(nc-c.data)/nc<1e-5).all(), (np.abs(nc-c.data).max(), np.where(np.abs(nc-c.data)>1))
            tp = float(report[-1][-1])
            assert tp > rtp * 10**9, (tp, rtp)
            
        check(64*1, 60*1, 64*1, 0, 0, 0, 31)
        check(64*1, 60*1, 64*1, 0, 1, 0, 25)
        check(64*1, 60*1, 64*1, 1, 0, 0, 42)
        check(64*55, 60*55, 64*55, 0, 0, 0, 20)
        check(64*55, 60*55, 64*55, 1, 1, 0, 37)
        check(64*55, 60*55, 64*55, 0, 1, 0, 36)
        check(64*55+1, 60*55+1, 64*55, 0, 1, 0, 36)
        check(64*55+1, 60*55+1, 64*55+1, 0, 1, 0, 36)
        check(64*55+15, 60*55+15, 64*55+15, 0, 1, 0, 34) # TODO: 36
        check(64*16, 60*16, 64*16, 0, 1, 0, 35)
        check(64*55, 60*55, 64*55, 0, 1, 0, 36)

    
    @unittest.skipIf(skip_slow_test, "Skip slow test")
    def test_profile_fused_op_split3(self):
        # match v6

        n, m, k = 64*100, 60*100, 64*100
        
        a = jt.random([n,m,1])
        b = jt.random([1,m,k])
        # a = jt.ones([n,m,1])
        # b = jt.ones([1,m,k])
        a.data, b.data
        with performance_test_scope(
            0, 400000000000//(n*m*k),
            compile_options={
            "order0":0, "order1":0,"order2":0,
            "split0":64*4,"split1":60*4,"split2":64*4,
        
            "order3":0, "order4":1,"order5":1,
            "split3":64,"split4":60,"split5":64,
        
            "order6":0, "order7":1,"order8":1,
            "split6":16,"split7":6,"split8":16,
        
            "order9":0, "order10":0,"order11":0,
        
            "restride":1, "unroll":2,"vectorize":2
        }):
            c = (a*b).sum(1)
        c.data

        na = a.data.reshape((n,m))
        nb = b.data.reshape((m,k))
        nc = np.matmul(na,nb)
        
        assert (np.abs(nc-c.data)/nc<1e-5).all(), (np.abs(nc-c.data).max(), np.where(np.abs(nc-c.data)>1))

    @unittest.skipIf(skip_slow_test, "Skip slow test")
    def test_profile_fused_op_parallel(self):
        # match v6

        @retry(10)
        def check(n, m, k, cs, rs, pa, rtp):
            a = jt.random([n,m,1])
            b = jt.random([1,m,k])
            a.data, b.data
            with performance_test_scope(
                2, 30,
                compile_options = {
                "order0":0, "order1":0,"order2":0,
                "split0":64,"split1":60,"split2":64,
                "split3":16,"split4":6,"split5":16,
                "order3":0, "order4":1,"order5":1,
                "order6":0, "order7":0,"order8":0,
                "restride":rs,"vectorize":2,"unroll":2,
                "compile_shapes":cs, "parallel":pa
            }) as report:
                c = (a*b).sum(1)
            c.data

            na = a.data.reshape((n,m))
            nb = b.data.reshape((m,k))
            nc = np.matmul(na,nb)
            
            assert (np.abs(nc-c.data)/nc<1e-5).all(), (np.abs(nc-c.data).max(), np.where(np.abs(nc-c.data)>1))
            tp = float(report[-1][-1])
            assert tp > rtp * 10**9, (tp, rtp)
            
        check(64*16, 60*16, 64*16, 1, 1, 0, 35)
        check(64*16, 60*16, 64*16, 1, 1, 1, 60)
        check(64*16, 60*16, 64*16, 0, 1, 0, 35)
        check(64*16, 60*16, 64*16, 0, 1, 1, 60)
        check(64*16+5, 60*16+5, 64*16+5, 0, 1, 1, 50)

if __name__ == "__main__":
    unittest.main()